use super::super::types::{OverlayDesc, OverlaySection};
use tool_helper::{Error, format_if_error, Output};
use std::io::Write;

const OVERLAY_DEFAULT_START:&'static str = "__planschi_start"; // < will probably be "__boot_loader_start" later

macro_rules! section_end_name_template {
    () => {
      "__{}_end"  
    };
}

pub fn write(output: &mut Output, overlay_desc: &OverlayDesc) -> Result<(), Error> {
    fn create_heap_base(overlay_desc: &OverlayDesc) -> String {
        let (first_entry, additional_slots_iter) = (&overlay_desc[0].name, overlay_desc.iter().skip(1));

        let mut heap_start = format!(concat!("MAX({}, ",  section_end_name_template!(), ")"), OVERLAY_DEFAULT_START, first_entry);
        for slot in additional_slots_iter {
            heap_start = format!(concat!("MAX(", section_end_name_template!(), ", {})"), slot.name, heap_start);
        }
        heap_start
    }

    if overlay_desc.is_empty() {
        write_heap_base(output, OVERLAY_DEFAULT_START)?;
    }

    else {
        write_heap_base(output, &create_heap_base(&overlay_desc))?;
    }

    let mut slot_start_adr = format!(section_end_name_template!(), "bss");

    writeln!(output, "")?;
    writeln!(output, "SECTIONS {{")?;
        writeln!(output, "\t/DISCARD/ : {{ *(.comment*) *(.gnu.*) *(.mdebug*) *(.debug_*) *(.pdr)}}")?;
        for slot in overlay_desc {
            writeln!(output, "\t/*{}*/", slot.name)?;
            writeln!(output, "\tOVERLAY {} : NOCROSSREFS SUBALIGN(4) {{", slot_start_adr)?;
                write_section(output, &slot.sections, false)?;
            writeln!(output, "\t}}")?;
            writeln!(output, "\t. = ALIGN(4);")?;
            writeln!(output, concat!("\t", section_end_name_template!(), " = .;"), slot.name)?;
            writeln!(output, "")?;

            slot_start_adr = format!(section_end_name_template!(), slot.name);
        }
    writeln!(output, "}}")?;
    return Ok(());
}

fn write_heap_base(output: &mut Output, heap_base: &str) -> Result<(), Error> {
    format_if_error!(writeln!(output, "__heap_base = {};", heap_base), "Writing default LD Script failed with: {error_text}")?;
    Ok(())
}

fn write_section(output: &mut Output, sections: &Vec<OverlaySection>, add_end_name: bool) -> Result<(), Error> {
    for section in sections {
        writeln!(output, "\t\t.{} {{", section.name)?;
            writeln!(output, "\t\t\t__{}_start = .;", section.name)?;
            writeln!(output, "\t\t\t__{}_lbas = .;", section.name)?;
            section.file_pattern.iter().try_for_each(|patr| writeln!(output, "\t\t\tKEEP({}(.header.lbas))", patr))?;
            writeln!(output, "\t\t\t__{}_ctor = .;", section.name)?;

            for (section_type, keep) in [(".text.startup._GLOBAL__*", false), (".ctors", false), (".text.*", false), (".rodata*", false), (".sdata*", false), (".data*", false), (".sbss*", false), (".bss*", false), (".keep.dummy", true)] {
                section.file_pattern.iter().try_for_each(|patr| {
                    let section_entry = format!("{}({})", patr, section_type);
                    
                    write!(output, "\t\t\t")?;
                    if keep {
                        writeln!(output, "KEEP({})", section_entry)
                    }
                    
                    else {
                        writeln!(output, "{}", section_entry)
                    }
                })?;
            }

            if add_end_name {
                writeln!(output, concat!("\t\t\t", section_end_name_template!(), " = .;"), section.name)?;
            }
        writeln!(output, "\t\t}}")?;
    }

    Ok(())
}